# Copyright (c) the SPDX tools authors
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import io
import json
import ntpath
import os
import posixpath
import re
from collections import OrderedDict
from typing import List

import xmltodict
import yaml

import spdx
from spdx import utils
from spdx.relationship import Relationship
from spdx.utils import NoAssert

test_data_dir = os.path.join(os.path.dirname(__file__), 'data')


def get_test_loc(test_path, test_data_dir=test_data_dir, debug=False, exists=True):
    """
    Given a `test_path` relative to the `test_data_dir` directory, return the
    location to a test file or directory for this path. No copy is done.
    """
    if debug:
        import inspect
        caller = inspect.stack()[1][3]
        print('\nget_test_loc,%(caller)s,"%(test_path)s","%(test_data_dir)s"' % locals())

    assert test_path
    assert test_data_dir

    if not os.path.exists(test_data_dir):
        raise IOError("[Errno 2] No such directory: test_data_dir not found:"
                      " '%(test_data_dir)s'" % locals())

    tpath = to_os_native_path(test_path)
    test_loc = os.path.abspath(os.path.join(test_data_dir, tpath))

    if exists and not os.path.exists(test_loc):
        raise IOError("[Errno 2] No such file or directory: "
                      "test_path not found: '%(test_loc)s'" % locals())

    return test_loc


def to_os_native_path(path):
    """
    Normalize a path to use the native OS path separator.
    """
    path = path.replace(posixpath.sep, os.path.sep)
    path = path.replace(ntpath.sep, os.path.sep)
    path = path.rstrip(os.path.sep)
    return path


def strip_variable_text(rdf_text):
    """
    Return rdf_text stripped from variable parts such as rdf nodeids
    """

    replace_nid = re.compile('rdf:nodeID="[^\"]*"').sub
    rdf_text = replace_nid('', rdf_text)

    replace_creation = re.compile('<ns1:creationInfo>.*</ns1:creationInfo>', re.DOTALL).sub
    rdf_text = replace_creation('', rdf_text)

    replace_pcc = re.compile('<ns1:packageVerificationCode>.*</ns1:packageVerificationCode>', re.DOTALL).sub
    rdf_text = replace_pcc('', rdf_text)
    return rdf_text


def load_and_clean_rdf(location):
    """
    Return plain Python nested data for the SPDX RDF file at location suitable
    for comparison. The file content is cleaned from variable parts such as
    dates, generated UUIDs and versions

    NOTE: we use plain dicts to avoid ordering issues in XML. the SPDX tool and
    lxml do not seem to return a consistent ordering that is needed for tests.
    """
    with io.open(location, encoding='utf-8') as l:
        content = l.read()
    content = strip_variable_text(content)
    data = xmltodict.parse(content, dict_constructor=dict)
    return sort_nested(data)


def sort_nested(data):
    """
    Return a new dict with any nested list sorted recursively.
    """
    if isinstance(data, dict):
        new_data = {}
        for k, v in data.items():
            if isinstance(v, list):
                v = sort_nested(v)
            if isinstance(v, dict):
                v = sort_nested(v)
            new_data[k] = v
        return new_data
    elif isinstance(data, list):
        new_data = []
        for v in sorted(data, key=lambda x: json.dumps(x, sort_keys=True)):
            if isinstance(v, list):
                v = sort_nested(v)
            if isinstance(v, dict):
                v = sort_nested(v)
            new_data.append(v)
        return new_data


def check_rdf_scan(expected_file, result_file, regen=False):
    """
    Check that expected and result_file are equal.
    Both are paths to SPDX RDF XML files, UTF-8 encoded.
    """
    result = load_and_clean_rdf(result_file)
    if regen:
        expected = result
        with io.open(expected_file, 'w', encoding='utf-8') as o:
            json.dump(expected, o, indent=2)
    else:
        with io.open(expected_file, 'r', encoding='utf-8') as i:
            expected = sort_nested(json.load(i))
    assert result == expected


def load_and_clean_tv(location):
    """
    Return a mapping for the SPDX TV file at location suitable for comparison.
    The file content is cleaned from variable parts such as dates, generated
    UUIDs and versions
    """
    with io.open(location, encoding='utf-8') as l:
        content = l.read()
    content = [l for l in content.splitlines(False)
               if l and l.strip() and not l.startswith(('Creator: ', 'Created: ',))]
    return '\n'.join(content)


def check_tv_scan(expected_file, result_file, regen=False):
    """
    Check that expected and result_file are equal.
    Both are paths to plain SPDX tv text files, UTF-8 encoded.
    """
    result = load_and_clean_tv(result_file)
    if regen:
        with io.open(expected_file, 'w') as o:
            o.write(result)

    expected = load_and_clean_tv(expected_file)
    assert result == expected


def load_and_clean_json(location):
    """
    Return plain Python nested data for the SPDX JSON file at location suitable
    for comparison. The file content is cleaned from variable parts such as
    dates, generated UUIDs and versions
    """
    with io.open(location, encoding='utf-8') as l:
        content = l.read()
    data = json.loads(content)

    if 'creationInfo' in data:
        del (data['creationInfo'])

    return sort_nested(data)


def check_json_scan(expected_file, result_file, regen=False):
    """
    Check that expected_file and result_file are equal.
    Both are paths to SPDX JSON files, UTF-8 encoded.
    """
    result = load_and_clean_json(result_file)
    if regen:
        with io.open(expected_file, 'w', encoding='utf-8') as o:
            o.write(result)

    expected = load_and_clean_json(expected_file)
    assert result == expected


def load_and_clean_yaml(location):
    """
    Return plain Python nested data for the SPDX YAML file at location suitable
    for comparison. The file content is cleaned from variable parts such as
    dates, generated UUIDs and versions
    """
    with io.open(location, encoding='utf-8') as l:
        content = l.read()
    data = yaml.safe_load(content)

    if 'creationInfo' in data:
        del (data['creationInfo'])

    return sort_nested(data)


def check_yaml_scan(expected_file, result_file, regen=False):
    """
    Check that expected_file and result_file are equal.
    Both are paths to SPDX YAML files, UTF-8 encoded.
    """
    result = load_and_clean_yaml(result_file)
    if regen:
        with io.open(expected_file, 'w', encoding='utf-8') as o:
            o.write(result)

    expected = load_and_clean_yaml(expected_file)
    assert result == expected


def load_and_clean_xml(location):
    """
    Return plain Python nested data for the SPDX XML file at location suitable
    for comparison. The file content is cleaned from variable parts such as
    dates, generated UUIDs and versions
    """
    with io.open(location, encoding='utf-8') as l:
        content = l.read()
    data = xmltodict.parse(content, encoding='utf-8')

    if 'creationInfo' in data['Document']:
        del (data['Document']['creationInfo'])

    return sort_nested(data)


def check_xml_scan(expected_file, result_file, regen=False):
    """
    Check that expected_file and result_file are equal.
    Both are paths to SPDX XML files, UTF-8 encoded.
    """
    result = load_and_clean_xml(result_file)
    if regen:
        with io.open(expected_file, 'w', encoding='utf-8') as o:
            o.write(result)

    expected = load_and_clean_xml(expected_file)
    assert result == expected


class TestParserUtils(object):
    """
    Helper class to represent SPDX Document models as Python types after parsing
    to be compared to expected data from a JSON file.
    """

    @classmethod
    def license_to_dict(cls, license_var):
        """
        Represents spdx.license.License, spdx.license.LicenseConjunction or
        spdx.license.LicenseDisjunction as a Python dictionary
        """
        CONJ_SEP = re.compile(' AND | and ')
        DISJ_SEP = re.compile(' OR | or ')
        if license_var is None:
            return None
        license_dict = OrderedDict()

        if isinstance(license_var, NoAssert):
            license_dict['type'] = 'Single'
            license_dict['identifier'] = 'NOASSERTION'
            license_dict['name'] = 'NOASSERTION'
            return license_dict
        elif isinstance(license_var, spdx.license.LicenseConjunction):
            license_dict['type'] = 'Conjunction'
            sep_regex = CONJ_SEP
        elif isinstance(license_var, spdx.license.LicenseDisjunction):
            license_dict['type'] = 'Disjunction'
            sep_regex = DISJ_SEP
        else:
            license_dict['type'] = 'Single'
            license_dict['identifier'] = license_var.identifier
            license_dict['name'] = license_var.full_name
            return license_dict

        license_dict['identifier'] = sorted(sep_regex.split(license_var.identifier))
        license_dict['name'] = sorted(sep_regex.split(license_var.full_name))

        return license_dict

    @classmethod
    def version_to_dict(cls, version):
        """
        Represents spdx.version.Version as a Python dictionary
        """
        return OrderedDict([
            ('major', int(version.major)),
            ('minor', int(version.minor))
        ])

    @classmethod
    def entity_to_dict(cls, entity):
        """
        Represents spdx.creationInfo.Creator subclasses as a dictionary
        """
        if entity is None:
            return None
        if isinstance(entity, utils.NoAssert):
            return entity.to_value()
        entity_dict = OrderedDict(name=entity.name)

        if isinstance(entity, spdx.creationinfo.Tool):
            entity_dict['type'] = 'Tool'
            return entity_dict

        entity_dict['email'] = entity.email
        entity_dict['type'] = 'Person'

        if isinstance(entity, spdx.creationinfo.Organization):
            entity_dict['type'] = 'Organization'
            return entity_dict

        return entity_dict

    @classmethod
    def checksum_to_dict(cls, checksum):
        """
        Represents spdx.checksum.Algorithm as a Python dictionary
        """
        if checksum is None:
            return None
        return OrderedDict([
            ('identifier', checksum.identifier.name),
            ('value', checksum.value)])

    @classmethod
    def package_to_dict(cls, package):
        """
        Represents spdx.package.Package as a Python dictionary
        """
        lics_from_files = []
        if package.are_files_analyzed:
            lics_from_files = sorted(package.licenses_from_files, key=lambda lic: lic.identifier)
        package_dict = OrderedDict([
            ('id', package.spdx_id),
            ('name', package.name),
            ('packageFileName', package.file_name),
            ('summary', package.summary),
            ('description', package.description),
            ('versionInfo', package.version),
            ('sourceInfo', package.source_info),
            ('downloadLocation', str(package.download_location)),
            ('homepage', package.homepage),
            ('originator', cls.entity_to_dict(package.originator)),
            ('supplier', cls.entity_to_dict(package.supplier)),
            ('licenseConcluded', cls.license_to_dict(package.conc_lics)),
            ('licenseDeclared', cls.license_to_dict(package.license_declared)),
            ('copyrightText', str(package.cr_text)),
            ('licenseComment', package.license_comment),
            ('checksums', [cls.checksum_to_dict(checksum) for checksum in package.checksums.values()]),
            ('licenseInfoFromFiles', [cls.license_to_dict(lic) for lic in lics_from_files]),
            ('verificationCode', OrderedDict([
                ('value', package.verif_code),
                ('excludedFilesNames', sorted(package.verif_exc_files))])
             )
        ])

        if package.built_date:
            package_dict['builtDate'] = utils.datetime_iso_format(package.built_date)

        if package.release_date:
            package_dict['releaseDate'] = utils.datetime_iso_format(package.release_date)

        if package.valid_until_date:
            package_dict['validUntilDate'] = utils.datetime_iso_format(package.valid_until_date)

        if package.primary_package_purpose:
            package_dict['primaryPackagePurpose'] = package.primary_package_purpose.name.replace("_", "-")

        return package_dict

    @classmethod
    def files_to_list(cls, files):
        """
        Represents a list of spdx.file.File as a Python list of dictionaries
        """
        files_list = []

        for file in files:
            lics_in_files = sorted(file.licenses_in_file, key=lambda lic: lic.identifier)
            chk_sums = []
            for checksum in file.checksums.values():
                chk_sums.append(cls.checksum_to_dict(checksum))

            file_dict = OrderedDict([
                ('id', file.spdx_id),
                ('fileName', file.name),
                ('fileTypes', [file_type.name for file_type in file.file_types]),
                ('comment', file.comment),
                ('licenseConcluded', cls.license_to_dict(file.conc_lics)),
                ('copyrightText', file.copyright),
                ('licenseComment', file.license_comment),
                ('notice', file.notice),
                ('checksums', chk_sums),
                ('licenseInfoInFiles', [cls.license_to_dict(lic) for lic in lics_in_files]),
                ('contributors', sorted(file.contributors)),
                ('dependencies', sorted(file.dependencies)),
                ('artifactOfProjectName', file.artifact_of_project_name),
                ('artifactOfProjectHome', file.artifact_of_project_home),
                ('artifactOfProjectURI', file.artifact_of_project_uri),
            ])
            files_list.append(file_dict)

        return files_list

    @classmethod
    def ext_document_references_to_list(cls, ext_doc_refs):
        """
        Represents a list of spdx.document.ExternalDocumentRef as a Python list of dictionaries
        """
        ext_doc_refs_list = []

        for ext_doc_ref in ext_doc_refs:
            ext_doc_ref_dict = OrderedDict([
                ('externalDocumentId', ext_doc_ref.external_document_id),
                ('spdxDocument', ext_doc_ref.spdx_document_uri),
                ('checksum', cls.checksum_to_dict(ext_doc_ref.checksum)),
            ])
            ext_doc_refs_list.append(ext_doc_ref_dict)

        return ext_doc_refs_list

    @classmethod
    def extracted_licenses_to_list(cls, extracted_licenses):
        """
        Represents a list of spdx.document.ExtractedLicense as a Python list of dictionaries
        """
        extracted_licenses_list = []

        for extracted_license in extracted_licenses:
            extracted_license_dict = OrderedDict([
                ('name', extracted_license.full_name),
                ('identifier', extracted_license.identifier),
                ('text', extracted_license.text),
                ('comment', extracted_license.comment),
                ('cross_refs', sorted(extracted_license.cross_ref)),
            ])
            if extracted_license_dict not in extracted_licenses_list:
                extracted_licenses_list.append(extracted_license_dict)

        return extracted_licenses_list

    @classmethod
    def annotations_to_list(cls, annotations):
        """
        Represents a list of spdx.annotation.Annotation as a Python list of dictionaries
        """
        annotations_list = []

        for annotation in annotations:
            annotation_dict = OrderedDict([
                ('id', annotation.spdx_id),
                ('comment', annotation.comment),
                ('type', annotation.annotation_type),
                ('annotator', cls.entity_to_dict(annotation.annotator)),
                ('date', utils.datetime_iso_format(annotation.annotation_date)),
            ])
            annotations_list.append(annotation_dict)

        return annotations_list

    @classmethod
    def reviews_to_list(cls, reviews):
        """
        Represents a list of spdx.review.Review as a Python list of dictionaries
        """
        reviews_list = []

        for review in reviews:
            review_dict = OrderedDict([
                ('comment', review.comment),
                ('reviewer', cls.entity_to_dict(review.reviewer)),
                ('date', utils.datetime_iso_format(review.review_date))
            ])
            reviews_list.append(review_dict)

        return reviews_list

    @classmethod
    def snippets_to_list(cls, snippets):
        """
        Represents a list of spdx.snippet.Snippet as a Python list of dictionaries
        """
        snippets_list = []

        for snippet in snippets:
            lics_from_snippet = sorted(snippet.licenses_in_snippet, key=lambda lic: lic.identifier)
            snippet_dict = OrderedDict([
                ('id', snippet.spdx_id),
                ('name', snippet.name),
                ('comment', snippet.comment),
                ('copyrightText', snippet.copyright),
                ('licenseComments', snippet.license_comment),
                ('snippetFromFile', snippet.snip_from_file_spdxid),
                ('licenseConcluded', cls.license_to_dict(snippet.conc_lics)),
                ('licenseInfoInSnippets', [cls.license_to_dict(lic) for lic in lics_from_snippet]),
            ])
            snippets_list.append(snippet_dict)

        return snippets_list

    @classmethod
    def relationships_to_dict_list(cls, relationships: List[Relationship]) -> List[OrderedDict]:
        relationships_list = []
        for relationship in relationships:
            relationship_dict = OrderedDict([
                ('spdx_element_id', relationship.spdx_element_id),
                ('relationship_type', relationship.relationship_type),
                ('related_spdx_element', relationship.related_spdx_element)
            ])
            relationships_list.append(relationship_dict)

        return relationships_list

    @classmethod
    def to_dict(cls, doc):
        """
        Represents a SPDX Document (spdx.document.Document) as nested Python types
        """
        creators = sorted(doc.creation_info.creators, key=lambda c: c.name)
        return OrderedDict([
            ('id', doc.spdx_id),
            ('specVersion', cls.version_to_dict(doc.version)),
            ('documentNamespace', doc.namespace),
            ('name', doc.name),
            ('comment', doc.comment),
            ('dataLicense', cls.license_to_dict(doc.data_license)),
            ('licenseListVersion', cls.version_to_dict(doc.creation_info.license_list_version)),
            ('creators', [cls.entity_to_dict(creator) for creator in creators]),
            ('created', utils.datetime_iso_format(doc.creation_info.created)),
            ('creatorComment', doc.creation_info.comment),
            ('files', cls.files_to_list(sorted(doc.files))),
            ('packages', [cls.package_to_dict(p) for p in doc.packages]),
            ('externalDocumentRefs', cls.ext_document_references_to_list(sorted(doc.ext_document_references))),
            ('extractedLicenses', cls.extracted_licenses_to_list(sorted(doc.extracted_licenses))),
            ('annotations', cls.annotations_to_list(sorted(doc.annotations))),
            ('reviews', cls.reviews_to_list(sorted(doc.reviews))),
            ('snippets', cls.snippets_to_list(sorted(doc.snippet))),
            ('relationships', cls.relationships_to_dict_list(doc.relationships))
        ])



# Copyright (c) 2021 spdx tool contributors
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spdx.parsers.loggers import ErrorMessages


def test_error_message_context():
    messages = ErrorMessages()
    messages.push_context("package1")
    messages.append("missing value: {0} {bar}", "foo", bar="bar")
    messages.append("missing key")
    messages.pop_context()
    messages.append("lone message")
    messages.push_context("package2")
    messages.push_context("file1")
    messages.append("more message")
    assert messages.messages == [
        "package1: missing value: foo bar",
        "package1: missing key",
        "lone message",
        "package2: file1: more message",
    ]



# Copyright (c) 2014 Ahmed H. Ismail
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil
import tempfile
import unittest
from datetime import datetime
from unittest import TestCase

from spdx.checksum import Checksum, ChecksumAlgorithm
from spdx.config import LICENSE_MAP, EXCEPTION_MAP
from spdx.creationinfo import Tool
from spdx.document import Document, ExternalDocumentRef
from spdx.license import License
from spdx.file import File, FileType
from spdx.package import Package, PackagePurpose
from spdx.parsers.loggers import ErrorMessages
from spdx.relationship import Relationship, RelationshipType
from spdx.utils import NoAssert
from spdx.version import Version

from tests import utils_test


class TestVersion(TestCase):
    maxDiff = None

    def test_creation(self):
        v = Version(major=2, minor=1)
        assert v.major == 2
        assert v.minor == 1

    def test_comparison(self):
        v1 = Version(major=1, minor=2)
        v2 = Version(major=2, minor=1)
        assert v1 != v2
        assert v1 < v2
        assert v1 <= v2
        assert v2 > v1
        assert v2 >= v1
        v3 = Version(major=1, minor=2)
        assert v3 == v1
        assert not v1 < v3
        assert v1 <= v3


class TestDocument(TestCase):
    maxDiff = None

    def test_creation(self):
        document = Document(
            version=Version(major=2, minor=1),
            data_license=License(full_name='Academic Free License v1.1',
                                identifier='AFL-1.1')
        )
        document.add_ext_document_reference(
            ExternalDocumentRef('DocumentRef-spdx-tool-2.1',
                                'https://spdx.org/spdxdocs/spdx-tools-v2.1-3F2504E0-4F89-41D3-9A0C-0305E82C3301',
                                Checksum(ChecksumAlgorithm.SHA1, 'SOME-SHA1'))
        )
        assert document.comment is None
        assert document.version == Version(2, 1)
        assert document.data_license.identifier == 'AFL-1.1'
        assert document.ext_document_references[-1].external_document_id == 'DocumentRef-spdx-tool-2.1'
        assert document.ext_document_references[-1].spdx_document_uri == 'https://spdx.org/spdxdocs/spdx-tools-v2.1-3F2504E0-4F89-41D3-9A0C-0305E82C3301'
        assert document.ext_document_references[-1].checksum.identifier.name == 'SHA1'
        assert document.ext_document_references[-1].checksum.value == 'SOME-SHA1'

    def test_document_validate_failures_returns_informative_messages(self):
        doc = Document(Version(2, 1), License.from_identifier('CC0-1.0'),
                       'Sample_Document-V2.1', spdx_id='SPDXRef-DOCUMENT',
                       namespace='https://spdx.org/spdxdocs/spdx-example-444504E0-4F89-41D3-9A0C-0305E82C3301')
        pack = doc.package = Package('some/path', NoAssert())
        pack.set_checksum(Checksum(ChecksumAlgorithm.SHA256, 'SOME-SHA256'))
        file1 = File('./some/path/tofile')
        file1.name = './some/path/tofile'
        file1.spdx_id = 'SPDXRef-File'
        file1.set_checksum(Checksum(ChecksumAlgorithm.SHA1, 'SOME-SHA1'))
        lic1 = License.from_identifier('LGPL-2.1-only')
        file1.add_lics(lic1)
        pack.add_lics_from_file(lic1)
        messages = ErrorMessages()
        messages = doc.validate(messages)
        expected = ['Sample_Document-V2.1: Creation info missing created date.',
                    'Sample_Document-V2.1: No creators defined, must have at least one.',
                    'Sample_Document-V2.1: some/path: Package download_location can not be None.']
        assert sorted(expected) == sorted(messages)

    def test_document_is_valid_when_using_or_later_licenses(self):
        doc = Document(Version(2, 1), License.from_identifier('CC0-1.0'),
                       'Sample_Document-V2.1', spdx_id='SPDXRef-DOCUMENT',
                       namespace='https://spdx.org/spdxdocs/spdx-example-444504E0-4F89-41D3-9A0C-0305E82C3301')
        doc.creation_info.add_creator(Tool('ScanCode'))
        doc.creation_info.set_created_now()

        package = doc.package = Package(name='some/path', download_location=NoAssert())
        package.spdx_id = 'SPDXRef-Package'
        package.cr_text = 'Some copyright'
        package.set_checksum(Checksum(ChecksumAlgorithm.SHA1, 'SOME-SHA1'))
        package.verif_code = 'SOME code'
        package.license_declared = NoAssert()
        package.conc_lics = NoAssert()

        file1 = File('./some/path/tofile')
        file1.name = './some/path/tofile'
        file1.spdx_id = 'SPDXRef-File'
        file1.file_types = [FileType.OTHER]
        file1.set_checksum(Checksum(ChecksumAlgorithm.SHA1, 'SOME-SHA1'))
        file1.conc_lics = NoAssert()
        file1.copyright = NoAssert()

        lic1 = License.from_identifier('LGPL-2.1-or-later')
        file1.add_lics(lic1)

        package.add_lics_from_file(lic1)
        doc.add_file(file1)
        relationship = create_relationship(package.spdx_id, RelationshipType.CONTAINS, file1.spdx_id)
        doc.add_relationship(relationship)
        messages = ErrorMessages()
        messages = doc.validate(messages)
        assert not messages

    def test_document_multiple_packages(self):
        doc = Document(Version(2, 1), License.from_identifier('CC0-1.0'),
                       'Sample_Document-V2.1', spdx_id='SPDXRef-DOCUMENT',
                       namespace='https://spdx.org/spdxdocs/spdx-example-444504E0-4F89-41D3-9A0C-0305E82C3301')
        doc.creation_info.add_creator(Tool('ScanCode'))
        doc.creation_info.set_created_now()

        package1 = Package(name='some/path1', download_location=NoAssert())
        package1.spdx_id = 'SPDXRef-Package1'
        package1.cr_text = 'Some copyright'
        package1.files_verified = False
        package1.license_declared = NoAssert()
        package1.conc_lics = NoAssert()
        doc.add_package(package1)

        package2 = Package(name='some/path2', download_location=NoAssert())
        package2.spdx_id = 'SPDXRef-Package2'
        package2.cr_text = 'Some copyright'
        package2.files_verified = False
        package2.license_declared = NoAssert()
        package2.conc_lics = NoAssert()
        doc.add_package(package2)

        assert len(doc.packages) == 2

    def test_document_without_packages(self):
        doc = Document(Version(2, 1), License.from_identifier("CC0-1.0"),
                       'Sample_Document_V2.1', spdx_id='SPDXRef-DOCUMENT',
                       namespace='https://spdx.org/spdxdocs/spdx-example-444504E0-4F89-41D3-9A0C-0305E82C3301')
        doc.creation_info.add_creator(Tool('ScanCode'))
        doc.creation_info.set_created_now()

        messages = doc.validate()
        assert len(messages.messages) == 0

class TestWriters(TestCase):
    maxDiff = None

    def _get_lgpl_doc(self, or_later=False):
        doc = Document(Version(2, 1), License.from_identifier('CC0-1.0'),
                       'Sample_Document-V2.1', spdx_id='SPDXRef-DOCUMENT',
                       namespace='https://spdx.org/spdxdocs/spdx-example-444504E0-4F89-41D3-9A0C-0305E82C3301')
        doc.creation_info.add_creator(Tool('ScanCode'))
        doc.creation_info.set_created_now()

        package = doc.package = Package(name='some/path', download_location=NoAssert())
        package.spdx_id = 'SPDXRef-Package'
        package.cr_text = 'Some copyright'
        package.verif_code = 'SOME code'
        package.set_checksum(Checksum(ChecksumAlgorithm.SHA1, 'SOME-SHA1'))
        package.set_checksum(Checksum(ChecksumAlgorithm.SHA256, 'SOME-SHA256'))
        package.license_declared = NoAssert()
        package.conc_lics = NoAssert()
        package.primary_package_purpose = PackagePurpose.FILE
        package.release_date = datetime(2021, 1, 1, 12, 0, 0)
        package.built_date = datetime(2021, 1, 1, 12, 0, 0)
        package.valid_until_date = datetime(2022, 1, 1, 12, 0, 0)


        file1 = File('./some/path/tofile')
        file1.name = './some/path/tofile'
        file1.spdx_id = 'SPDXRef-File'
        file1.set_checksum(Checksum(ChecksumAlgorithm.SHA1, 'SOME-SHA1'))
        file1.set_checksum(Checksum(ChecksumAlgorithm.SHA256, 'SOME-SHA256'))
        file1.conc_lics = NoAssert()
        file1.copyright = NoAssert()
        file1.file_types = [FileType.OTHER, FileType.SOURCE]

        lic1 = License.from_identifier('LGPL-2.1-only')
        if or_later:
            lic1 = License.from_identifier('LGPL-2.1-or-later')

        file1.add_lics(lic1)
        doc.add_file(file1)

        package.add_lics_from_file(lic1)
        relationship = create_relationship(package.spdx_id, RelationshipType.CONTAINS, file1.spdx_id)
        doc.add_relationship(relationship)
        relationship = create_relationship(doc.spdx_id, RelationshipType.DESCRIBES, package.spdx_id)
        doc.add_relationship(relationship)
        return doc

    def _get_lgpl_multi_package_doc(self, or_later=False):
        doc = Document(Version(2, 1), License.from_identifier('CC0-1.0'),
                       'Sample_Document-V2.1', spdx_id='SPDXRef-DOCUMENT',
                       namespace='https://spdx.org/spdxdocs/spdx-example-444504E0-4F89-41D3-9A0C-0305E82C3301')
        doc.creation_info.add_creator(Tool('ScanCode'))
        doc.creation_info.set_created_now()

        # This package does not have files analyzed
        package1 = Package(name='some/path1', download_location=NoAssert())
        package1.spdx_id = 'SPDXRef-Package1'
        package1.cr_text = 'Some copyright'
        package1.files_analyzed = False
        package1.license_declared = NoAssert()
        package1.conc_lics = NoAssert()
        doc.add_package(package1)

        # This one does, which is the default
        package2 = Package(name='some/path2', download_location=NoAssert())
        package2.spdx_id = 'SPDXRef-Package2'
        package2.cr_text = 'Some copyright'
        package2.license_declared = NoAssert()
        package2.conc_lics = NoAssert()
        package2.verif_code = 'SOME code'

        # This one does, but per recommendation, we choose to make the
        # verification code optional in this library
        package3 = Package(name='some/path3', download_location=NoAssert())
        package3.spdx_id = 'SPDXRef-Package3'
        package3.cr_text = 'Some copyright'
        package3.license_declared = NoAssert()
        package3.conc_lics = NoAssert()

        file1 = File('./some/path/tofile')
        file1.name = './some/path/tofile'
        file1.spdx_id = 'SPDXRef-File'
        file1.set_checksum(Checksum(ChecksumAlgorithm.SHA1, 'SOME-SHA1'))
        file1.conc_lics = NoAssert()
        file1.copyright = NoAssert()

        lic1 = License.from_identifier('LGPL-2.1-only')
        if or_later:
            lic1 = License.from_identifier('LGPL-2.1-or-later')

        file1.add_lics(lic1)

        package2.add_lics_from_file(lic1)
        package3.add_lics_from_file(lic1)

        doc.add_package(package2)
        doc.add_package(package3)
        doc.add_file(file1)

        relationship = create_relationship(doc.spdx_id, RelationshipType.DESCRIBES, package1.spdx_id)
        doc.add_relationship(relationship)
        relationship = create_relationship(doc.spdx_id, RelationshipType.DESCRIBES, package2.spdx_id)
        doc.add_relationship(relationship)
        relationship = create_relationship(doc.spdx_id, RelationshipType.DESCRIBES, package3.spdx_id)
        doc.add_relationship(relationship)
        relationship = create_relationship(package2.spdx_id, RelationshipType.CONTAINS, file1.spdx_id)
        doc.add_relationship(relationship)
        relationship = create_relationship(package3.spdx_id, RelationshipType.CONTAINS, file1.spdx_id)
        doc.add_relationship(relationship)

        return doc

    def test_write_document_rdf_with_validate(self):
        from spdx.writers.rdf import write_document
        doc = self._get_lgpl_doc()
        temp_dir = ''
        try:
            temp_dir = tempfile.mkdtemp(prefix='test_spdx')
            result_file = os.path.join(temp_dir, 'spdx-simple.rdf')
            with open(result_file, 'wb') as output:
                write_document(doc, output, validate=True)

            expected_file = utils_test.get_test_loc(
                'doc_write/rdf-simple.json',
                test_data_dir=utils_test.test_data_dir)

            utils_test.check_rdf_scan(expected_file, result_file, regen=False)
        finally:
            if temp_dir and os.path.exists(temp_dir):
                shutil.rmtree(temp_dir)

    def test_write_document_rdf_with_or_later_with_validate(self):
        from spdx.writers.rdf import write_document
        doc = self._get_lgpl_doc(or_later=True)

        temp_dir = ''
        try:
            temp_dir = tempfile.mkdtemp(prefix='test_spdx')
            result_file = os.path.join(temp_dir, 'spdx-simple-plus.rdf')

            # test proper!
            with open(result_file, 'wb') as output:
                write_document(doc, output, validate=True)

            expected_file = utils_test.get_test_loc(
                'doc_write/rdf-simple-plus.json',
                test_data_dir=utils_test.test_data_dir)

            utils_test.check_rdf_scan(expected_file, result_file, regen=False)
        finally:
            if temp_dir and os.path.exists(temp_dir):
                shutil.rmtree(temp_dir)

    def test_write_document_tv_with_validate(self):
        from spdx.writers.tagvalue import write_document
        doc = self._get_lgpl_doc()

        temp_dir = ''
        try:
            temp_dir = tempfile.mkdtemp(prefix='test_spdx')
            result_file = os.path.join(temp_dir, 'spdx-simple.tv')
            with open(result_file, 'w') as output:
                write_document(doc, output, validate=True)

            expected_file = utils_test.get_test_loc(
                'doc_write/tv-simple.tv',
                test_data_dir=utils_test.test_data_dir)

            utils_test.check_tv_scan(expected_file, result_file, regen=False)
        finally:
            if temp_dir and os.path.exists(temp_dir):
                shutil.rmtree(temp_dir)

    def test_write_document_tv_with_or_later_with_validate(self):
        from spdx.writers.tagvalue import write_document

        doc = self._get_lgpl_doc(or_later=True)

        temp_dir = ''
        try:
            temp_dir = tempfile.mkdtemp(prefix='test_spdx')
            result_file = os.path.join(temp_dir, 'spdx-simple-plus.tv')

            # test proper!
            with open(result_file, 'w') as output:
                write_document(doc, output, validate=True)

            expected_file = utils_test.get_test_loc(
                'doc_write/tv-simple-plus.tv',
                test_data_dir=utils_test.test_data_dir)

            utils_test.check_tv_scan(expected_file, result_file, regen=False)
        finally:
            if temp_dir and os.path.exists(temp_dir):
                shutil.rmtree(temp_dir)

    def test_write_document_json_with_validate(self):
        from spdx.writers.json import write_document
        doc = self._get_lgpl_doc()

        temp_dir = ''
        try:
            temp_dir = tempfile.mkdtemp(prefix='test_spdx')
            result_file = os.path.join(temp_dir, 'spdx-simple.json')
            with open(result_file, 'w') as output:
                write_document(doc, output, validate=True)

            expected_file = utils_test.get_test_loc(
                'doc_write/json-simple.json',
                test_data_dir=utils_test.test_data_dir)

            utils_test.check_json_scan(expected_file, result_file, regen=False)
        finally:
            if temp_dir and os.path.exists(temp_dir):
                shutil.rmtree(temp_dir)

    def test_write_document_json_with_or_later_with_validate(self):
        from spdx.writers.json import write_document
        doc = self._get_lgpl_doc(or_later=True)

        temp_dir = ''
        try:
            temp_dir = tempfile.mkdtemp(prefix='test_spdx')
            result_file = os.path.join(temp_dir, 'spdx-simple-plus.json')
            with open(result_file, 'w') as output:
                write_document(doc, output, validate=True)

            expected_file = utils_test.get_test_loc(
                'doc_write/json-simple-plus.json',
                test_data_dir=utils_test.test_data_dir)

            utils_test.check_json_scan(expected_file, result_file, regen=False)
        finally:
            if temp_dir and os.path.exists(temp_dir):
                shutil.rmtree(temp_dir)

    def test_write_document_json_multi_package_with_validate(self):
        from spdx.writers.json import write_document
        doc = self._get_lgpl_multi_package_doc(or_later=True)

        temp_dir = ''
        try:
            temp_dir = tempfile.mkdtemp(prefix='test_spdx')
            result_file = os.path.join(temp_dir, 'spdx-simple-multi-package.json')
            with open(result_file, 'w') as output:
                write_document(doc, output, validate=True)

            expected_file = utils_test.get_test_loc(
                'doc_write/json-simple-multi-package.json',
                test_data_dir=utils_test.test_data_dir)

            utils_test.check_json_scan(expected_file, result_file, regen=False)
        finally:
            if temp_dir and os.path.exists(temp_dir):
                shutil.rmtree(temp_dir)


    def test_write_document_yaml_with_validate(self):
        from spdx.writers.yaml import write_document
        doc = self._get_lgpl_doc()

        temp_dir = ''
        try:
            temp_dir = tempfile.mkdtemp(prefix='test_spdx')
            result_file = os.path.join(temp_dir, 'spdx-simple.yaml')
            with open(result_file, 'w') as output:
                write_document(doc, output, validate=True)

            expected_file = utils_test.get_test_loc(
                'doc_write/yaml-simple.yaml',
                test_data_dir=utils_test.test_data_dir)

            utils_test.check_yaml_scan(expected_file, result_file, regen=False)
        finally:
            if temp_dir and os.path.exists(temp_dir):
                shutil.rmtree(temp_dir)

    def test_write_document_yaml_with_or_later_with_validate(self):
        from spdx.writers.yaml import write_document
        doc = self._get_lgpl_doc(or_later=True)

        temp_dir = ''
        try:
            temp_dir = tempfile.mkdtemp(prefix='test_spdx')
            result_file = os.path.join(temp_dir, 'spdx-simple-plus.yaml')
            with open(result_file, 'w') as output:
                write_document(doc, output, validate=True)

            expected_file = utils_test.get_test_loc(
                'doc_write/yaml-simple-plus.yaml',
                test_data_dir=utils_test.test_data_dir)

            utils_test.check_yaml_scan(expected_file, result_file, regen=False)
        finally:
            if temp_dir and os.path.exists(temp_dir):
                shutil.rmtree(temp_dir)

    def test_write_document_yaml_multi_package_with_validate(self):
        from spdx.writers.yaml import write_document
        doc = self._get_lgpl_multi_package_doc(or_later=True)

        temp_dir = ''
        try:
            temp_dir = tempfile.mkdtemp(prefix='test_spdx')
            result_file = os.path.join(temp_dir, 'spdx-simple-multi-package.yaml')
            with open(result_file, 'w') as output:
                write_document(doc, output, validate=True)

            expected_file = utils_test.get_test_loc(
                'doc_write/yaml-simple-multi-package.yaml',
                test_data_dir=utils_test.test_data_dir)

            utils_test.check_yaml_scan(expected_file, result_file, regen=False)
        finally:
            if temp_dir and os.path.exists(temp_dir):
                shutil.rmtree(temp_dir)

    def test_write_document_xml_with_validate(self):
        from spdx.writers.xml import write_document
        doc = self._get_lgpl_doc()

        temp_dir = ''
        try:
            temp_dir = tempfile.mkdtemp(prefix='test_spdx')
            result_file = os.path.join(temp_dir, 'spdx-simple.xml')
            with open(result_file, 'w') as output:
                write_document(doc, output, validate=True)

            expected_file = utils_test.get_test_loc(
                'doc_write/xml-simple.xml',
                test_data_dir=utils_test.test_data_dir)

            utils_test.check_xml_scan(expected_file, result_file, regen=False)
        finally:
            if temp_dir and os.path.exists(temp_dir):
                shutil.rmtree(temp_dir)

    def test_write_document_xml_with_or_later_with_validate(self):
        from spdx.writers.xml import write_document
        doc = self._get_lgpl_doc(or_later=True)

        temp_dir = ''
        try:
            temp_dir = tempfile.mkdtemp(prefix='test_spdx')
            result_file = os.path.join(temp_dir, 'spdx-simple-plus.xml')
            with open(result_file, 'w') as output:
                write_document(doc, output, validate=True)

            expected_file = utils_test.get_test_loc(
                'doc_write/xml-simple-plus.xml',
                test_data_dir=utils_test.test_data_dir)

            utils_test.check_xml_scan(expected_file, result_file, regen=False)
        finally:
            if temp_dir and os.path.exists(temp_dir):
                shutil.rmtree(temp_dir)

    def test_write_document_xml_multi_package_with_validate(self):
        from spdx.writers.xml import write_document
        doc = self._get_lgpl_multi_package_doc(or_later=True)

        temp_dir = ''
        try:
            temp_dir = tempfile.mkdtemp(prefix='test_spdx')
            result_file = os.path.join(temp_dir, 'spdx-simple-multi-package.xml')
            with open(result_file, 'w') as output:
                write_document(doc, output, validate=True)

            expected_file = utils_test.get_test_loc(
                'doc_write/xml-simple-multi-package.xml',
                test_data_dir=utils_test.test_data_dir)

            utils_test.check_xml_scan(expected_file, result_file, regen=False)
        finally:
            if temp_dir and os.path.exists(temp_dir):
                shutil.rmtree(temp_dir)

    def _get_mini_doc(self,):
        doc = Document(Version(2, 1), License.from_identifier('CC0-1.0'))
        doc.creation_info.set_created_now()

        return doc

    def test_write_document_tv_mini(self):
        from spdx.writers.tagvalue import write_document
        doc = self._get_mini_doc()

        temp_dir = ''
        try:
            temp_dir = tempfile.mkdtemp(prefix='test_spdx')
            result_file = os.path.join(temp_dir, 'spdx-simple.tv')
            with open(result_file, 'w') as output:
                write_document(doc, output, validate=False)
            expected_file = utils_test.get_test_loc('doc_write/tv-mini.tv')
            utils_test.check_tv_scan(expected_file, result_file, regen=False)
        finally:
            if temp_dir and os.path.exists(temp_dir):
                shutil.rmtree(temp_dir)

    def test_write_document_rdf_mini(self):
        from spdx.writers.rdf import write_document
        doc = self._get_mini_doc()
        temp_dir = ''
        try:
            temp_dir = tempfile.mkdtemp(prefix='test_spdx')
            result_file = os.path.join(temp_dir, 'spdx-simple.rdf')
            with open(result_file, 'wb') as output:
                write_document(doc, output, validate=False)
            expected_file = utils_test.get_test_loc('doc_write/rdf-mini.json')
            utils_test.check_rdf_scan(expected_file, result_file, regen=False)
        finally:
            if temp_dir and os.path.exists(temp_dir):
                shutil.rmtree(temp_dir)


def create_relationship(spdx_element_id: str, relationship_type: RelationshipType, related_spdx_element: str) -> Relationship:
    return Relationship(spdx_element_id + " " + relationship_type.name + " " + related_spdx_element)


class TestLicense(TestCase):
    maxDiff = None

    def test_url(self):
        lic = License(full_name='Apache License 1.0', identifier='Apache-1.0')
        assert lic.url == 'http://spdx.org/licenses/Apache-1.0'

    def test_license_list(self):
        assert LICENSE_MAP['Aladdin Free Public License'] == 'Aladdin'
        assert LICENSE_MAP['Aladdin'] == 'Aladdin Free Public License'
        assert LICENSE_MAP['MIT License'] == 'MIT'
        assert LICENSE_MAP['MIT'] == 'MIT License'
        assert LICENSE_MAP['BSD 4-Clause "Original" or "Old" License'] == 'BSD-4-Clause'
        assert LICENSE_MAP['BSD-4-Clause'] == 'BSD 4-Clause "Original" or "Old" License'

    def test_from_full_name(self):
        mit = License.from_full_name('MIT License')
        assert mit.identifier == 'MIT'
        assert mit.url == 'http://spdx.org/licenses/MIT'

    def test_from_identifier(self):
        mit = License.from_identifier('MIT')
        assert mit.full_name == 'MIT License'
        assert mit.url == 'http://spdx.org/licenses/MIT'


class TestException(TestCase):

    def test_exception_list(self):
        assert EXCEPTION_MAP['Linux Syscall Note'] == 'Linux-syscall-note'
        assert EXCEPTION_MAP['Linux-syscall-note'] == 'Linux Syscall Note'
        assert EXCEPTION_MAP['GCC Runtime Library exception 3.1'] == 'GCC-exception-3.1'
        assert EXCEPTION_MAP['GCC-exception-3.1'] == 'GCC Runtime Library exception 3.1'



# Copyright (c) 2014 Ahmed H. Ismail
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re
from enum import Enum, auto


class ChecksumAlgorithm(Enum):
    SHA1 = auto()
    SHA224 = auto()
    SHA256 = auto()
    SHA384 = auto()
    SHA512 = auto()
    SHA3_256 = auto()
    SHA3_384 = auto()
    SHA3_512 = auto()
    BLAKE2B_256 = auto()
    BLAKE2B_384 = auto()
    BLAKE2B_512 = auto()
    BLAKE3 = auto()
    MD2 = auto()
    MD4 = auto()
    MD5 = auto()
    MD6 = auto()
    ADLER32 = auto()

    def algorithm_to_rdf_representation(self) -> str:
        if self.name.startswith("BLAKE2B"):
            return "checksumAlgorithm_" + self.name.replace("_", "").lower()
        else:
            return "checksumAlgorithm_" + self.name.lower()

    @classmethod
    def checksum_from_rdf(cls, identifier: str) -> 'ChecksumAlgorithm':
        identifier = identifier.split('_', 1)[-1].upper()
        blake_checksum = re.compile(r"^(BLAKE2B)(256|384|512)$", re.UNICODE)
        match = blake_checksum.match(identifier)
        if match:
            identifier = match[1] + '_' + match[2]
        if identifier not in ChecksumAlgorithm.__members__:
            raise ValueError(f"Invalid algorithm for checksum: {identifier}")
        return ChecksumAlgorithm[identifier]

    @classmethod
    def checksum_algorithm_from_string(cls, identifier: str) -> 'ChecksumAlgorithm':
        identifier = identifier.replace("-", "_").upper()
        if identifier not in ChecksumAlgorithm.__members__:
            raise ValueError(f"Invalid algorithm for checksum: {identifier}")
        return ChecksumAlgorithm[identifier]


class Checksum(object):
    """Generic checksum algorithm."""

    def __init__(self, identifier: ChecksumAlgorithm, value: str):
        self.identifier = identifier
        self.value = value

    def __eq__(self, other) -> bool:
        if not isinstance(other, Checksum):
            return False
        return self.identifier == other.identifier and self.value == other.value

    @classmethod
    def checksum_from_string(cls, value: str) -> 'Checksum':
        CHECKSUM_RE = re.compile("(ADLER32|BLAKE2b-256|BLAKE2b-384|BLAKE2b-512|BLAKE3|MD2|MD4|MD5|MD6|" \
                                 "SHA1|SHA224|SHA256|SHA384|SHA512|SHA3-256|SHA3-384|SHA3-512):\\s*([a-f0-9]+)")
        match = CHECKSUM_RE.match(value)
        if match is None or match.group(1) is None or match.group(2) is None:
            raise ValueError(f"Invalid checksum: {value}")
        identifier = ChecksumAlgorithm.checksum_algorithm_from_string(match.group(1))
        return Checksum(identifier=identifier, value=match.group(2))

    def to_tv(self) -> str:
        algorithm_name: str = self.identifier.name
        # Convert underscores to dashes, and other Blake2b-specific casing rules
        if "_" in algorithm_name:
            algorithm_name = CHECKSUM_ALGORITHM_TO_TV.get(algorithm_name)
            if algorithm_name is None:
                raise ValueError(f"Missing conversion rule for converting {self.identifier.name} to tag-value string")
        return "{0}: {1}".format(algorithm_name, self.value)


CHECKSUM_ALGORITHM_TO_TV = {
    ChecksumAlgorithm.BLAKE2B_256.name: "BLAKE2b-256",
    ChecksumAlgorithm.BLAKE2B_384.name: "BLAKE2b-384",
    ChecksumAlgorithm.BLAKE2B_512.name: "BLAKE2b-512",
    ChecksumAlgorithm.SHA3_256.name: "SHA3-256",
    ChecksumAlgorithm.SHA3_384.name: "SHA3-384",
    ChecksumAlgorithm.SHA3_512.name: "SHA3-512"
}



# Copyright (c) 2018 Yash M. Nisar
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from functools import total_ordering

from spdx.utils import datetime_iso_format


@total_ordering
class Annotation(object):

    """
    Document annotation information.
    Fields:
    - annotator: Person, Organization or tool that has commented on a file,
    package, or the entire document. Conditional (Mandatory, one), if there
    is an Annotation.
    - annotation_date: To identify when the comment was made. Conditional
    (Mandatory, one), if there is an Annotation. Type: datetime.
    - comment: Annotation comment. Conditional (Mandatory, one), if there is
    an Annotation. Type: str.
    - annotation_type: Annotation type. Conditional (Mandatory, one), if there is an
    Annotation. Type: str.
    - spdx_id: Uniquely identify the element in an SPDX document which is being
    referenced. Conditional (Mandatory, one), if there is an Annotation.
    Type: str.
    """

    def __init__(
        self,
        annotator=None,
        annotation_date=None,
        comment=None,
        annotation_type=None,
        spdx_id=None,
    ):
        self.annotator = annotator
        self.annotation_date = annotation_date
        self.comment = comment
        self.annotation_type = annotation_type
        self.spdx_id = spdx_id

    def __eq__(self, other):
        return (
            isinstance(other, Annotation)
            and self.annotator == other.annotator
            and self.annotation_date == other.annotation_date
            and self.comment == other.comment
        )

    def __lt__(self, other):
        return (self.annotator, self.annotation_date, self.comment) < (
            other.annotator,
            other.annotation_date,
            other.comment,
        )

    def set_annotation_date_now(self):
        self.annotation_date = datetime.utcnow()

    @property
    def annotation_date_iso_format(self):
        return datetime_iso_format(self.annotation_date)

    @property
    def has_comment(self):
        return self.comment is not None

    def validate(self, messages):
        """
        Check that all the fields are valid.
        Appends any error messages to messages parameter shall be a ErrorMessages.
        """
        self.validate_annotator(messages)
        self.validate_annotation_date(messages)
        self.validate_annotation_type(messages)
        self.validate_spdx_id(messages)

        return messages

    def validate_annotator(self, messages):
        if self.annotator is None:
            messages.append("Annotation missing annotator.")

    def validate_annotation_date(self, messages):
        if self.annotation_date is None:
            messages.append("Annotation missing annotation date.")

    def validate_annotation_type(self, messages):
        if self.annotation_type is None:
            messages.append("Annotation missing annotation type.")

    def validate_spdx_id(self, messages):
        if self.spdx_id is None:
            messages.append("Annotation missing SPDX Identifier Reference.")



# Copyright (c) 2014 Ahmed H. Ismail
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re

from ply import lex


class Lexer(object):
    reserved = {
        # Top level fields
        "SPDXVersion": "DOC_VERSION",
        "DataLicense": "DOC_LICENSE",
        "DocumentName": "DOC_NAME",
        "SPDXID": "SPDX_ID",
        "DocumentComment": "DOC_COMMENT",
        "DocumentNamespace": "DOC_NAMESPACE",
        "ExternalDocumentRef": "EXT_DOC_REF",
        # Creation info
        "Creator": "CREATOR",
        "Created": "CREATED",
        "CreatorComment": "CREATOR_COMMENT",
        "LicenseListVersion": "LIC_LIST_VER",
        # Review info
        "Reviewer": "REVIEWER",
        "ReviewDate": "REVIEW_DATE",
        "ReviewComment": "REVIEW_COMMENT",
        # Annotation info
        "Annotator": "ANNOTATOR",
        "AnnotationDate": "ANNOTATION_DATE",
        "AnnotationComment": "ANNOTATION_COMMENT",
        "AnnotationType": "ANNOTATION_TYPE",
        "SPDXREF": "ANNOTATION_SPDX_ID",
        # Relationships
        "Relationship": "RELATIONSHIP",
        "RelationshipComment": "RELATIONSHIP_COMMENT",
        # Package Fields
        "PackageName": "PKG_NAME",
        "PackageVersion": "PKG_VERSION",
        "PackageDownloadLocation": "PKG_DOWN",
        "FilesAnalyzed": "PKG_FILES_ANALYZED",
        "PackageSummary": "PKG_SUM",
        "PackageSourceInfo": "PKG_SRC_INFO",
        "PackageFileName": "PKG_FILE_NAME",
        "PackageSupplier": "PKG_SUPPL",
        "PackageOriginator": "PKG_ORIG",
        "PackageChecksum": "PKG_CHKSUM",
        "PackageVerificationCode": "PKG_VERF_CODE",
        "PackageDescription": "PKG_DESC",
        "PackageComment": "PKG_COMMENT",
        "PackageLicenseDeclared": "PKG_LICS_DECL",
        "PackageLicenseConcluded": "PKG_LICS_CONC",
        "PackageLicenseInfoFromFiles": "PKG_LICS_FFILE",
        "PackageLicenseComments": "PKG_LICS_COMMENT",
        "PackageCopyrightText": "PKG_CPY_TEXT",
        "PackageHomePage": "PKG_HOME",
        "ExternalRef": "PKG_EXT_REF",
        "ExternalRefComment": "PKG_EXT_REF_COMMENT",
        "PackageAttributionText": "PKG_ATTRIBUTION_TEXT",
        "PrimaryPackagePurpose": "PRIMARY_PACKAGE_PURPOSE",
        "BuiltDate": "BUILT_DATE",
        "ReleaseDate": "RELEASE_DATE",
        "ValidUntilDate": "VALID_UNTIL_DATE",
        # Files
        "FileName": "FILE_NAME",
        "FileType": "FILE_TYPE",
        "FileChecksum": "FILE_CHKSUM",
        "LicenseConcluded": "FILE_LICS_CONC",
        "LicenseInfoInFile": "FILE_LICS_INFO",
        "FileCopyrightText": "FILE_CR_TEXT",
        "LicenseComments": "FILE_LICS_COMMENT",
        "FileComment": "FILE_COMMENT",
        "FileNotice": "FILE_NOTICE",
        "FileContributor": "FILE_CONTRIB",
        "FileDependency": "FILE_DEP",
        "ArtifactOfProjectName": "ART_PRJ_NAME",
        "ArtifactOfProjectHomePage": "ART_PRJ_HOME",
        "ArtifactOfProjectURI": "ART_PRJ_URI",
        "FileAttributionText": "FILE_ATTRIBUTION_TEXT",
        # License
        "LicenseID": "LICS_ID",
        "ExtractedText": "LICS_TEXT",
        "LicenseName": "LICS_NAME",
        "LicenseCrossReference": "LICS_CRS_REF",
        "LicenseComment": "LICS_COMMENT",
        # Snippet
        "SnippetSPDXID": "SNIPPET_SPDX_ID",
        "SnippetName": "SNIPPET_NAME",
        "SnippetComment": "SNIPPET_COMMENT",
        "SnippetCopyrightText": "SNIPPET_CR_TEXT",
        "SnippetLicenseComments": "SNIPPET_LICS_COMMENT",
        "SnippetFromFileSPDXID": "SNIPPET_FILE_SPDXID",
        "SnippetLicenseConcluded": "SNIPPET_LICS_CONC",
        "LicenseInfoInSnippet": "SNIPPET_LICS_INFO",
        "SnippetAttributionText": "SNIPPET_ATTRIBUTION_TEXT",
        "SnippetByteRange": "SNIPPET_BYTE_RANGE",
        "SnippetLineRange": "SNIPPET_LINE_RANGE",
        # Common
        "NOASSERTION": "NO_ASSERT",
        "UNKNOWN": "UN_KNOWN",
        "NONE": "NONE",
        "SOURCE": "SOURCE",
        "BINARY": "BINARY",
        "ARCHIVE": "ARCHIVE",
        "APPLICATION": "APPLICATION",
        "AUDIO": "AUDIO",
        "IMAGE": "IMAGE",
        "TEXT": "FILETYPE_TEXT",
        "VIDEO": "VIDEO",
        "DOCUMENTATION": "DOCUMENTATION",
        "SPDX": "SPDX",
        "OTHER": "OTHER",
        "REVIEW": "REVIEW",
        "FRAMEWORK": "FRAMEWORK",
        "LIBRARY": "LIBRARY",
        "CONTAINER": "CONTAINER",
        "OPERATING-SYSTEM": "OPERATING_SYSTEM",
        "DEVICE": "DEVICE",
        "FIRMWARE": "FIRMWARE",
        "FILE": "FILE",
        "INSTALL": "INSTALL"
    }
    states = (("text", "exclusive"),)

    tokens = [
        "TEXT",
        "TOOL_VALUE",
        "UNKNOWN_TAG",
        "ORG_VALUE",
        "PERSON_VALUE",
        "DATE",
        "LINE",
        "CHKSUM",
        "DOC_REF_ID",
        "DOC_URI",
        "EXT_DOC_REF_CHKSUM",
    ] + list(reserved.values())

    def t_text(self, t):
        r":\s*<text>"
        t.lexer.text_start = t.lexer.lexpos - len("<text>")
        t.lexer.begin("text")

    def t_text_end(self, t):
        r"</text>\s*"
        t.type = "TEXT"
        t.value = t.lexer.lexdata[t.lexer.text_start: t.lexer.lexpos]
        t.lexer.lineno += t.value.count("\n")
        t.value = t.value.strip()
        t.lexer.begin("INITIAL")
        return t

    def t_text_any(self, t):
        r".|\n"
        pass

    def t_text_error(self, t):
        print("Lexer error in text state")

    def t_CHKSUM(self, t):
        r":\s*(ADLER32|BLAKE2b-256|BLAKE2b-384|BLAKE2b-512|BLAKE3|MD2|MD4|MD5|MD6|" \
        "SHA1|SHA224|SHA256|SHA384|SHA512|SHA3-256|SHA3-384|SHA3-512):\s*([a-f0-9]*)"
        t.value = t.value[1:].strip()
        return t


    def t_DOC_REF_ID(self, t):
        r":\s*DocumentRef-([A-Za-z0-9\+\.\-]+)"
        t.value = t.value[1:].strip()
        return t

    def t_DOC_URI(self, t):
        r"\s*((ht|f)tps?:\/\/\S*)"
        t.value = t.value.strip()
        return t

    def t_EXT_DOC_REF_CHKSUM(self, t):
        r"\s*SHA1:\s*[a-f0-9]{40,40}"
        t.value = t.value[1:].strip()
        return t

    def t_TOOL_VALUE(self, t):
        r":\s*Tool:.+"
        t.value = t.value[1:].strip()
        return t

    def t_ORG_VALUE(self, t):
        r":\s*Organization:.+"
        t.value = t.value[1:].strip()
        return t

    def t_PERSON_VALUE(self, t):
        r":\s*Person:.+"
        t.value = t.value[1:].strip()
        return t

    def t_DATE(self, t):
        r":\s*\d\d\d\d-\d\d-\d\dT\d\d:\d\d:\d\dZ"
        t.value = t.value[1:].strip()
        return t

    def t_KEYWORD_AS_TAG(self, t):
        r"[a-zA-Z]+"
        t.type = self.reserved.get(t.value, "UNKNOWN_TAG")
        t.value = t.value.strip()
        return t

    def t_LINE_OR_KEYWORD_VALUE(self, t):
        r":.+"
        t.value = t.value[1:].strip()
        if t.value in self.reserved.keys():
            t.type = self.reserved[t.value]
            return t
        else:
            t.type = "LINE"
        return t

    def t_comment(self, t):
        r"\#.*"
        pass

    def t_newline(self, t):
        r"\n+"
        t.lexer.lineno += len(t.value)

    def t_whitespace(self, t):
        r"[ \t]+"
        pass

    def build(self, **kwargs):
        self.lexer = lex.lex(module=self, **kwargs)

    def token(self):
        return self.lexer.token()

    def input(self, data):
        self.lexer.input(data)

    def t_error(self, t):
        t.lexer.skip(1)
        t.value = "Lexer error"
        return t



# Copyright (c) Xavier Figueroa
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import yaml

from spdx.writers.tagvalue import InvalidDocumentError
from spdx.writers.jsonyamlxml import Writer
from spdx.parsers.loggers import ErrorMessages


def write_document(document, out, validate=True):

    if validate:
        messages = ErrorMessages()
        messages = document.validate(messages)
        if messages:
            raise InvalidDocumentError(messages)

    writer = Writer(document)
    document_object = writer.create_document()

    yaml.safe_dump(document_object, out, indent=2, explicit_start=True, encoding='utf-8')
